/* CGV/solution/parts/cgvp01.c */

/****************************************************************
 *                                                              *
 * This file has been written as a sample solution to an        *
 * exercise in a course given at the High Performance           *
 * Computing Centre Stuttgart (HLRS).                           *
 * It is made freely available with the understanding that      *
 * every copy of this file must include this header and that    *
 * HLRS takes no responsibility for the use of the enclosed     *
 * teaching material.                                           *
 *                                                              *
 * Authors: Dr. Rolf Rabenseifner                               *
 *          Stephan Scheiderer                                  *
 *                                                              *
 * Contact: rabenseifner@hlrs.de                                *
 *                                                              *
 ****************************************************************/
                                                           /*EOH*/
/*>>>>>>>>>>>>>>>>>>>>>> begin of task 01 <<<<<<<<<<<<<<<<<<<<<<*/

/********************** vector routines ************************/

/*------------------------------
 * void alloc_vec
 *
 * RETURNS:
 *   the memory address of the allocated vector
 *
 * TASK:
 *   allocate memory (sizeof(doudle) * (physical size)) for a vector 
 *
 * GLOBAL DATA:
 *   IN int chunksize : size of the physical area this process is to calculate
 * USED FUNCTIONS:
 *   APPL:
 *     safe_malloc 
 *   C-lib:
 *     memset
 */
vector_chunk alloc_vec()
{
#ifndef WITH_MEMOPS
    int i;
#endif
    vector_chunk newVec = (vector_chunk) safe_malloc(sizeof(double) * chunksize);
#ifdef WITH_MEMOPS
    memset(newVec, 0, sizeof(double) * chunksize);
#else
    for (i = 0; i < chunksize; i++) {
        newVec[i] = 0;
    }
#endif
    return newVec;
}

/*------------------------------
 * void dupl_vec
 *   IN  vector_chunk v1      : vector 1 (to copy from)
 *   OUT vector_chunk v2      : vector 2 (to copy to)
 *
 * TASK:
 *   do a vector chunk duplication: v2 := v1
 *   easy solution: memcpy (see '> man memcpy')
 *
 * GLOBAL DATA:
 *   IN int chunksize : size of the physical area this process is to calculate
 * USED FUNCTIONS:
 *   C-lib:
 *     memcpy
 */
void dupl_vec(vector_chunk v1, vector_chunk v2)
{
#ifdef WITH_MEMOPS
    memcpy(v2, v1, sizeof(double) * chunksize);
#else
    int i;
    for (i = 0; i < chunksize; i++) {
        v2[i] = v1[i];
    }
#endif
}


/*-------------------------------------------
 * void add_vec_vec
 *  IN  vector_chunk v1      : vector 1
 *  IN  double       lambda  : factor lambda
 *  IN  vector_chunk v2      : vector 2
 *  OUT vector_chunk v3      : vector 3
 *
 * RETURNS:
 *   --
 *
 * TASK:
 *   calculate:   v3 := v1 + lambda * v2
 *
 * GLOBAL DATA:
 *   IN int chunksize : size of the physical area this process is to calculate
 *
 * USED FUNCTIONS:
 *   --
 */
void add_vec_vec(vector_chunk restrict v1, double lambda, vector_chunk restrict v2,
                 vector_chunk restrict /* out */ v3)
{
    int i;

    /* addition of two vectors, second vector is multiplied by a scalar */
    for (i = 0; i < chunksize; i++) {
        v3[i] = v1[i] + lambda * v2[i];
    }

    /* for the MFLOPS calculation */
    flops += chunksize * 2;
}

/*------------------------------
 * double dot_product_vec_vec
 *  IN  vector_chunk v1      : vector 1
 *  IN  vector_chunk v2      : vector 2
 *
 *  RETURNS:
 *    double                 : dot product
 *
 * TASK:
 *   calculate dot product for vector 1 and vector 2
 *
 * GLOBAL DATA:
 *   IN int chunksize : size of the physical area this process is to calculate
 *
 * USED FUNCTIONS:
 *   APPL:
 *     --
 *   MPI :
 *     MPI_Allreduce
 */
double dot_product_vec_vec(vector_chunk v1, vector_chunk v2)
{
    int i;
    double sum, global_sum;

    /* sum = dot product of vectors */
    sum = 0;
    for(i = 0; i < chunksize; i++) {
        sum += v1[i] * v2[i];
    }

/* global sum */
#ifdef serial
    global_sum = sum;
#else 
          MPI_Allreduce(&sum, &global_sum, 1, MPI_DOUBLE, MPI_SUM, comm_cart);
    flops += 1;
#endif

    /* for the MFLOPS calculation */
    flops += chunksize * 2;

    return global_sum;
}


/*------------------------------
 * double max_norm_vec
 *  IN  vector_chunk v1      : vector 1
 *
 *  RETURNS:
 *    double                 : maximum norm
 *
 * TASK:
 *   calculate (vector maximum norm) for vector 1
 *
 * GLOBAL DATA:
 *   IN int chunksize : size of the physical area this process is to calculate
 *
 * USED FUNCTIONS:
 *   APPL:
 *     -- 
 *   MPI :
 *     MPI_Allreduce
 */
double max_norm_vec(vector_chunk v1)
{
    int i;
    double maxabs, global_max;

    /* maxabs = maximum norm of the local vector */
          maxabs = 0;
          for(i = 0; i < chunksize; i++) {
              if ( v1[i] > maxabs) maxabs =  v1[i];
              if (-v1[i] > maxabs) maxabs = -v1[i];
          }

    /* global maximum */
#ifdef serial
          global_max = maxabs;
#else
          MPI_Allreduce(&maxabs, &global_max, 1, MPI_DOUBLE, MPI_MAX, comm_cart);
#endif

    /* no flops added, because we have only assignments and no operations */
          return global_max;
}

/*------------------------------
 * double sqr_norm_vec
 *  IN  vector_chunk v1      : vector 1
 *
 *  RETURNS:
 *    double            : square of norm
 *
 * TASK:
 *   calculate global (vector norm Euklid)^2 for vector 1
 *
 * GLOBAL DATA:
 *   IN int chunksize : size of the physical area this process is to calculate
 *
 * USED FUNCTIONS:
 *   APPL:
 *     -- 
 *   MPI :
 *     MPI_Allreduce
 */
double sqr_norm_vec(vector_chunk v1)
{
    int i;
    double sum, global_sum;

    /* sum = dot product of vector */
          sum = 0;
          for(i = 0; i < chunksize; i++) {
              sum += v1[i] * v1[i];
          }

    /* global sum */
#ifdef serial
          global_sum = sum;
#else
          MPI_Allreduce(&sum, &global_sum, 1, MPI_DOUBLE, MPI_SUM, comm_cart);
          flops += 1;
#endif

    /* for the MFLOPS calculation */
    flops += chunksize * 2;

          return global_sum;
}

/*>>>>>>>>>>>>>>>>>>>>>> -end- of task 01 <<<<<<<<<<<<<<<<<<<<<<*/
