/* CGV/solution/parts/cgvp04.c */

/****************************************************************
 *                                                              *
 * This file has been written as a sample solution to an        *
 * exercise in a course given at the High Performance           *
 * Computing Centre Stuttgart (HLRS).                           *
 * It is made freely available with the understanding that      *
 * every copy of this file must include this header and that    *
 * HLRS takes no responsibility for the use of the enclosed     *
 * teaching material.                                           *
 *                                                              *
 * Authors: Dr. Rolf Rabenseifner                               *
 *          Stephan Scheiderer                                  *
 *                                                              *
 * Contact: rabenseifner@hlrs.de                                *
 *                                                              *
 ****************************************************************/
                                                           /*EOH*/
/*>>>>>>>>>>>>>>>>>>>>>> begin of task 04 <<<<<<<<<<<<<<<<<<<<<<*/

#ifndef serial
/******************* domain decomposition **********************/

/*----------------------- 1-dim decomposition -----------------*/

/* //   Example: m=6, n=7, numprocs=2 */
/* // */
/* //       |       j=0     1     2     3     4     5     6 |           */
/* //   ----+-----------------------------------------------+-----------*/
/* //   i=0 |         0     1     2     3     4     5     6 |           */
/* //     1 |         7     8     9    10    11    12    13 | process 1 */
/* //     2 |        14    15    16    17    18    19    20 |           */
/* //   ----+-----------------------------------------------+-----------*/
/* //     3 |        21    22    23    24    25    26    27 |           */
/* //     4 |        28    29    30    31    32    33    34 | process 2 */
/* //     5 |        35    36    37    38    39    40    41 |           */
/* //   ----+-----------------------------------------------+-----------*/

/*------------------------------
 * Init_1dim
 *
 * TASK:
 *   Do the one dimensional domain decomposition as depicted above.
 *
 * USED GLOBAL DATA:
 *   OUT int start_i    : start row of the physical area this process is to calculate
 *   OUT int end1_i     : end row + 1 of the physical area this process is to calculate
 *   OUT int start_j    : start column of the physical area this process is to calculate
 *   OUT int end1_j     : end column + 1 of the physical area this process is to calculate
 *   IN/OUT int my_rank : the rank of the current process
 *   OUT MPI_Comm comm_cart : cartesian communicator
 *   IN  int numprocs   : number of used processors
 *
 * USED FUNCTIONS:
 *   C-lib:
 *     printf
 *   MPI:
 *     MPI_Cart_create, MPI_Comm_rank
 */
void Init_1dim()
{
    int dims[2];
    int periods[2];
    int slice_size;

    if ((print_level >= 1) && (my_rank == 0)) {
        printf("Using 1-dim domain decomposition, numprocs=%d\n", numprocs);
    }
    /* Init the virtual topology (1-dim) */
    dims[0] = numprocs;
    dims[1] = 1;
    periods[0] = FALSE;
    periods[1] = FALSE;
          MPI_Cart_create(MPI_COMM_WORLD, 2, dims, periods, FALSE, &comm_cart);
          MPI_Comm_rank(comm_cart, &my_rank);

    /* the size of each slice
     * If m is not a multiple of numprocs, we round up the slice size
     */
    slice_size = ((m - 1) / numprocs) + 1;

    /* the rows of the physical area are distributed in slices of equal size
     * (maybe except for the last slice) to each process
     */
    start_i = (my_rank * slice_size);
    if (start_i > m) start_i = m;

    end1_i = start_i + slice_size;
    if (end1_i > m) end1_i = m;

    /* we have the full vertical extent (all n columns) for each row */
    start_j = 0;
    end1_j = n;
}
#endif



#ifndef serial
/*----------------------- 2-dim decomposition -----------------*/

/* //   Example: m=5, n=7, m_procs=2, n_procs=2 */
/* // */
/* //       |           j=0     1     2     3  |  4     5     6            | */
/* //   ----+----------------------------------+---------------------------+ */
/* //   i=0 |             0     1     2     3  |  4     5     6            | */
/* //     1 | process 1   7     8     9    10  | 11    12    13  process 2 | */
/* //     2 |            14    15    16    17  | 18    19    20            | */
/* //   ----+----------------------------------+---------------------------+ */
/* //     3 | process 3  21    22    23    24  | 25    26    27  process 4 | */
/* //     4 |            28    29    30    31  | 32    33    34            | */
/* //   ----+------------------------------+-------------------------------+ */
/* // */

/*------------------------------
 * Init_2dim
 *
 * TASK:
 *   Do the two dimensional domain decomposition as depicted above.
 *
 * USED GLOBAL DATA:
 *   OUT int start_i   : start row of the physical area this process is to calculate
 *   OUT int end1_i    : end row + 1 of the physical area this process is to calculate
 *   OUT int start_j   : start column of the physical area this process is to calculate
 *   OUT int end1_j    : end column + 1 of the physical area this process is to calculate
 *   IN  int my_rank   : the rank of the current process
 *   IN  int m_procs   : vertical number of processors
 *   IN  int n_procs   : horizontal number of processors
 *   OUT MPI_Comm comm_cart : cartesian communicator
 *
 * USED FUNCTIONS:
 *   C-lib:
 *     printf
 *   MPI:
 *     MPI_Cart_create, MPI_Comm_rank
 */
void Init_2dim()
{
    int dims[2];
    int periods[2];
    int blocksize_i, blocksize_j;
    int startblock_i, startblock_j;

    if ((print_level >= 1) && (my_rank == 0)) {
        printf("Using 2-dim domain decomposition, m_procs=%d, n_procs=%d, numprocs=%d\n",
                m_procs, n_procs, numprocs);
    }

    /* Init the virtual topology (2-dims) */
    dims[0] = m_procs;
    dims[1] = n_procs;
    periods[0] = FALSE;
    periods[1] = FALSE;
          MPI_Cart_create(MPI_COMM_WORLD, 2, dims, periods, FALSE, &comm_cart);
          MPI_Comm_rank(comm_cart, &my_rank);

    if ((my_rank == 0) && ((m_procs * n_procs) != numprocs)) {
        printf("Number of processors does not fit! Use %d processors instead. "
               "Aborting...\n", (m_procs * n_procs));
        abrt();
    }

    /* the number of rows for each block (the last block might have
     * less than the first blocks) */
          blocksize_i = ((m - 1) / m_procs) + 1;

    /* the number of columns for each block (the last block might have
     * less than the first blocks) */
          blocksize_j = ((n - 1) / n_procs) + 1;

    /* calculate the start and end row and column */
          startblock_i = my_rank / n_procs;
          startblock_j = my_rank % n_procs;
          start_i = blocksize_i * startblock_i;
          end1_i = start_i + blocksize_i;
          if (end1_i > m) end1_i = m;
          start_j = blocksize_j * startblock_j;
          end1_j = start_j + blocksize_j;
          if (end1_j > n) end1_j = n;
}
#endif



/*------------------------------
 * Init_Decomp
 *
 * TASK:
 *   Do the domain decomposition.
 *
 * USED GLOBAL DATA:
 *   IN/OUT int rowsize     : number of data entries in a row of the local chunk
 *   IN/OUT int colsize     : number of data entries in a column of the local chunk
 *   IN/OUT int chunksize   : number of data entries in the local chunk
 *   IN     int decomp_dims : dimension of the domain decomposition 
 *   IN     int my_rank     : the rank of the current process
 *
 * USED FUNCTIONS:
 *   APPL:
 *     abrt
 *   C-lib:
 *     printf
 */
void Init_Decomp()
{
#ifndef serial
    switch (decomp_dims) {
        case 1:
            Init_1dim();
            break;
        case 2:
            Init_2dim();
            break;
        default:
            printf("Wrong decomp_dims=%d\n",decomp_dims);
            abrt();
    }
#else 
    start_i = 0;
    end1_i = m;
    start_j = 0;
    end1_j = n;
#endif 
    /* calculate the size of the rows, cols and the whole chunk */
    rowsize   = end1_j - start_j;
    colsize   = end1_i - start_i;
    chunksize = rowsize * colsize;

    if ((rowsize < 2) || (colsize < 2)) {
        printf("Each process must have at least 2 entries in each dimension.\n");
        printf("Decrease amount of processes or increase physical area size.\n");
        printf("Rank[%d]: rowsize = %d -- colsize = %d\n", my_rank, rowsize, colsize);
        printf("Aborting.\n");
        abrt();
    }
}

/*>>>>>>>>>>>>>>>>>>>>>> -end- of task 04 <<<<<<<<<<<<<<<<<<<<<<*/
